/* rfbproxy - a record/playback VNC proxy
 * Copyright (C) 2000  Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * This program works by recording everything the server says, and
 * then playing it back.  Yes, it _is_ that easy. ;-)
 *
 * I've found that hextile encoding gives the best results.
 *
 * See also: the TODO file.  */

/*
 * The file format is this:
 *
 * <capture file> ::- <data>
 * <data>         ::- <delay><byte-count><pad><raw-data><data>
 *                  | <delay><byte-count><pad><raw-data>
 * <delay>        ::- 32-bit number of milliseconds, big-endian
 * <byte-count>   ::- 16-bit number of bytes, big-endian
 * <pad>          ::- 16 bits (not used)
 * <raw-data>     ::- data received after a delay of <delay> since previous
 *                    packet, which is <byte-count> bytes in length, padded
 *                    to 32-bit boundary
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <getopt.h>
#include <netdb.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define VERSION "rfbproxy 0.9\n"

#define VNC_BASE 5900
#define DEFAULT_DISPLAY ":10"
#define DEFAULT_SERVER ":1"
#define BUFSIZE 65535

struct packet
{
	uint32_t ms;
	uint16_t len;
	uint16_t unused;
	/* 32-bit padded data */
};

static ssize_t read_traffic (int fd, void *buf, size_t len)
{
	static char *buffer;
	static int buffer_length;
	static int buffer_at;
	struct packet packet;

	if (!buffer)
		buffer = malloc (BUFSIZE);

	if (!buffer)
		return 0;

	if (fd < 0) {
		/* Reset the buffer */
		buffer_length = buffer_at = 0;
		return 0;
	}

	for (;;) {
		if (len < buffer_length - buffer_at) {
			memcpy (buf, buffer, len);
			buffer_at += len;
			return len;
		}

		if (buffer_length - buffer_at) {
			len = buffer_length - buffer_at;
			memcpy (buf, buffer, len);
			buffer_length = buffer_at = 0;
			return len;
		}

		if (read (fd, &packet, sizeof (struct packet)) <
		    sizeof (struct packet))
			return 0;

		buffer_length = ntohs (packet.len);
		buffer_at = 0;
		if (read (fd, buffer,
			  4 * ((buffer_length + 3) / 4)) < buffer_length)
			return 0;
	}
}

static void rewind_file (int fd)
{
	/* We have to rewind, and then skip authentication */
	char buf4[4];
	int state = 0;

	read_traffic (-1, NULL, 0);
	lseek (fd, 0, SEEK_SET);

	do {
		switch (state) {
		case 0: /* ProtocolVersion */
		{
			char skip[12];
			if (read_traffic (fd, skip, 12) < 12)
				goto too_short;
			state++;
			break;
		}
		case 1: /* Authentication */
		{
			char skip[20];
			if (read_traffic (fd, buf4, 4) < 4)
				goto too_short;
			if (buf4[3] == 2)
				if (read_traffic (fd, skip, 20) < 20)
					goto too_short;
			state++;
			if (read_traffic (fd, skip, 20) < 20)
				goto too_short;
			break;
		}
		case 2: /* ServerInitialisation */
		{
			char *skip;
			size_t len;
			if (read_traffic (fd, buf4, 4) < 4)
				goto too_short;
			len = ntohl (*(size_t*)buf4);
			skip = malloc (len);
			if (read_traffic (fd, skip, len) < len)
				goto too_short;
			free (skip);
			state++;
			break;
		}
		}
	} while (state < 3);
	return;

 too_short:
	fprintf (stderr, "Input has finished too early\n");
	exit (1);
}

static ssize_t do_write (int fd, const void *buf, size_t len)
{
	while (len) {
		ssize_t wrote = write (fd, buf, len);
		if (wrote < 0) {
			perror ("write");
			exit (1);
		}
		buf += wrote;
		len -= wrote;
	}
	return len;
}

static int record (const char *file, int clientr, int clientw,
		   struct sockaddr_in server_addr)
{
	FILE *f;
	int server = -1;
	struct timeval last_tv;
	int first = 1;
	char *buf = malloc (BUFSIZE);

	if (!buf) {
		fprintf (stderr, "Couldn't allocate buffer\n");
		exit (1);
	}

	f = fopen (file, "w");
	if (!f) {
		perror ("fopen");
		exit (1);
	}

	server = socket (PF_INET, SOCK_STREAM, IPPROTO_IP);
	if (server == -1) {
		perror ("socket");
		goto out;
	}

	if (connect (server, (struct sockaddr *) &server_addr,
		     sizeof (struct sockaddr_in))) {
		perror ("connect");
		goto out;
	}

	for (;;) {
		ssize_t bufs;
		struct timeval tv;
		struct timezone tz;
		fd_set rfds;
		FD_ZERO (&rfds);
		FD_SET (server, &rfds);
		FD_SET (clientr, &rfds);

		if (select(FD_SETSIZE, &rfds, NULL, NULL, NULL) < 0) {
			perror ("select");
			goto out;
		}

		if (FD_ISSET (server, &rfds)) {
			struct packet packet;			
			unsigned long diffms;
			struct timeval diff;

			gettimeofday (&tv, &tz);
			if (first) {
				first = 0;
				last_tv = tv;
			}
			bufs = read (server, buf, BUFSIZE);
			if (!bufs) break;
			do_write (clientw, buf, bufs);
			diff.tv_sec = tv.tv_sec - last_tv.tv_sec;
			diff.tv_usec = tv.tv_usec - last_tv.tv_usec;
			diffms = 1000 * diff.tv_sec;
			diffms += diff.tv_usec / 1000;
			packet.ms = htonl (diffms);
			packet.len = htons (bufs);
			fwrite (&packet, sizeof (packet), 1, f);
			fwrite (buf, 1, 4 * ((bufs + 3) / 4), f);
			last_tv = tv;
		}
		else {
			bufs = read (clientr, buf, BUFSIZE);
			if (!bufs) break;
			do_write (server, buf, bufs);
		}
	}
 out:
	free (buf);
	if (server != -1)
		close (server);
	fclose (f);
	return 0;
}

static int playback (const char *file, int clientr, int clientw, int loop)
{
	int fd = open (file, O_RDONLY);
	struct stat st;
	char *map, *buf;
	char *skip;
	off_t loop_offset = 0;

	if (fd < 0) {
		fprintf (stderr, "Couldn't open input file\n");
		return 1;
	}

	fstat (fd, &st);
	map = buf = mmap (NULL, st.st_size, PROT_READ | PROT_WRITE,
			  MAP_PRIVATE, fd, 0);
	close (fd);
	if (map == MAP_FAILED) {
		fprintf (stderr, "Couldn't map input file\n");
		return 1;
	}

	skip = malloc (BUFSIZE);
	if (!skip)
		goto out;

 loop:
	for (;;) {
		struct packet *packet;
		size_t len;
		struct timeval tv, deadline;
		struct timezone tz;
		unsigned long ms;
		fd_set rfds;
		FD_ZERO (&rfds);
		FD_SET (clientr, &rfds);

		if (map + st.st_size - buf < sizeof (struct packet))
			break;

		packet = (struct packet *) buf;
		ms = ntohl (packet->ms);
		tv.tv_sec = ms / 1000;
		tv.tv_usec = 1000 * (ms % 1000);
		len = ntohs (packet->len);

		if (map + st.st_size - buf < sizeof (struct packet) + len)
			break;

		gettimeofday (&deadline, &tz);
		deadline.tv_sec += tv.tv_sec;
		deadline.tv_usec += tv.tv_usec;
		if (deadline.tv_usec >= 1000000) {
			deadline.tv_usec -= 1000000;
			deadline.tv_sec++;
		}

		while (tv.tv_usec > 5000 &&
		       select(FD_SETSIZE, &rfds, NULL, NULL, &tv) != 0) {
			if (!read (clientr, skip, BUFSIZE))
				/* Connection closed. */
				goto out;

			/* Recalculate delay */
			gettimeofday (&tv, &tz);
			if (tv.tv_sec > deadline.tv_sec ||
			    (tv.tv_sec == deadline.tv_sec &&
			     tv.tv_usec >= deadline.tv_usec))
				/* Deadline already passed */
				break;

			tv.tv_sec = deadline.tv_sec - tv.tv_sec;
			tv.tv_usec = deadline.tv_usec - tv.tv_usec;
			if (tv.tv_usec < 0) {
				tv.tv_usec += 1000000;
				tv.tv_sec--;
			}
		}

		do_write (clientw, buf + sizeof (struct packet), len);
		buf += sizeof (struct packet) + 4 * ((len + 3) / 4);
	}

	if (loop) {
		if (!loop_offset) {
			fd = open (file, O_RDONLY);
			rewind_file (fd);
			loop_offset = lseek (fd, 0, SEEK_CUR);
			close (fd);
		}

		buf = map + loop_offset;
		goto loop;
	}

 out:
	free (skip);
	munmap (map, st.st_size);
	return 0;
}

static void version (void)
{
	fprintf (stderr, VERSION);
}

static void usage (const char *name)
{
	fprintf (stderr,
		 "usage: %s [OPTIONS] ACTION file\n"
		 "where OPTIONS are:\n\n"
		 " -c, --stdout  Use stdin and stdout for communications\n"
		 "               with the client. Useful in conjunction\n"
		 "               with inetd.\n"
		 " -l, --loop    (playback only) When file is finished, \n"
		 "               replay from first FrameBufferUpdate\n"
		 " --type=[screen|events]\n"
		 "               (record only) Capture either screen updates\n"
		 "               (\"screen\") or keyboard/mouse events (\"events\").\n"
		 "               The default is \"screen\".\n"
		 " :n            Occupy VNC display localhost:n (not valid with\n"
		 "               -c option).\n"
		 " --server=[server]:display\n"
		 "               (record only) Use specified VNC server. The\n"
		 "               default is \"" DEFAULT_SERVER "\".\n"
		 "\nACTION is one of:\n"
		 " -r file, --record=file\n"
		 "               Record RFB communications and store them in\n"
		 "               the file.\n"
		 " -p file, --playback=file\n"
		 "               Play back the RFB communications that were\n"
		 "               captured to the file\n",
		 name);
	exit (1);
}

int accept_connection (int port)
{
	int bound;
	int sock;
	struct sockaddr_in sin;
	int on = 1;
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = htons (VNC_BASE + port);
	bound = socket (AF_INET, SOCK_STREAM, IPPROTO_IP);
	if (bound < 0) {
		perror ("socket");
		exit (1);
	}
	setsockopt (bound, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on));
	if (bind (bound, (struct sockaddr *) &sin,
		  sizeof (struct sockaddr_in))) {
		perror ("bind");
		exit (1);
	}
	listen (bound, 1);
	sock = accept (bound, NULL, 0);
	close (bound);
	return sock;
}

int main (int argc, char *argv[])
{
	int use_stdout = 0;
	int loop = 0;
	const char *file = NULL;
	char action = '\0';
	char type = '\0';
	char *server = NULL;
	char *display = NULL;
	int clientr, clientw;
	struct sockaddr_in server_addr;

	/* Options */
	for (;;) {
		static struct option long_options[] = {
			{"playback", 1, 0, 'p'},
			{"record", 1, 0, 'r'},
			{"type", 1, 0, 't'},
			{"loop", 1, 0, 'l'},
			{"stdout", 0, 0, 'c'},
			{"server", 1, 0, 's'},
			{"help", 0, 0, 'h'},
			{"version", 0, 0, 'v'},
			{0, 0, 0, 0}
		};
		int l;
		int c = getopt_long (argc, argv, "p:r:cl",
				     long_options, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 'v':
			version ();
			exit (0);
		case 'h':
			version ();
			usage (argv[0]);
		case 'c':
			if (use_stdout)
				usage (argv[0]);
			use_stdout = 1;
			break;
		case 'l':
			if (loop)
				usage (argv[0]);
			loop = 1;
			break;
		case 't':
			if (type)
				usage (argv[0]);
			l = strlen (optarg);
			if (!strncmp (optarg, "screen", l))
				type = 's';
			else if (!strncmp (optarg, "events", l))
				type = 'e';
			else usage (argv[0]);
			break;
		case 's':
			if (server)
				usage (argv[0]);
			server = optarg;
			break;
		case 'p':
		case 'r':
			if (action)
				usage (argv[0]);
			action = c;
			file = optarg;
			break;
		}
	}

	while (optind < argc) {
		if (argv[optind][0] == ':') {
			if (display)
				usage (argv[0]);
			display = argv[optind];
		}
		optind++;
	}

	/* Invalid option combinations */
	if (!action ||
	    (loop && action != 'p') ||
	    (type && action != 'r') ||
	    (server && action != 'r') ||
	    (use_stdout && display))
		usage (argv[0]);

	/* Unimplemented bits */
	if (action == 'r' && type == 'e')
		fprintf (stderr,
			 "Warning: --type=events not implemented yet\n");

	/* Defaults */
	if (!server)
		server = strdup (DEFAULT_SERVER);
	if (!display)
		display = strdup (DEFAULT_DISPLAY);
	if (!type)
		type = 's';

	if (action == 'r') {
		int port;
		char *end;
		char *cl = strchr (server, ':');
		if (!cl)
			usage (argv[0]);
		*cl++ = '\0';
		port = VNC_BASE + strtoul (cl, &end, 10);
		if (cl == end)
			usage (argv[0]);
		server_addr.sin_family = AF_INET;
		if (!server[0])
			server_addr.sin_addr.s_addr = INADDR_ANY;
		else {
			struct hostent *hp;
			if ((server_addr.sin_addr.s_addr =
			     inet_addr (server)) == -1) {
				hp = gethostbyname (server);
				if (!hp) {
					perror (server);
					exit (1);
				}
				server_addr.sin_addr.s_addr = hp->h_addr[0];
			}
		}
		server_addr.sin_port = htons (port);
	}

	/* Get some file descriptors */
	if (use_stdout) {
		clientr = fileno (stdin);
		clientw = fileno (stdout);
	} else {
		unsigned long port;
		char *end;
		display++;
		port = strtoul (display, &end, 10);
		if (display == end)
			usage (argv[0]);
		clientr = clientw = accept_connection (port);
	}

	/* Do it */
	if (action == 'r') {
		record (file, clientr, clientw, server_addr);
	} else {
		playback (file, clientr, clientw, loop);
	}

	/* Clean up */
	if (!use_stdout)
		close (clientr);

	return 0;
}
